@@ -1,7 +1,7 @@
 /* vim: set et ts=8 sw=8: */
 /* gclue-service-client.c
  *
- * Copyright (C) 2013 Red Hat, Inc.
+ * Copyright 2013 Red Hat, Inc.
  *
  * Geoclue is free software; you can redistribute it and/or modify it under
  * the terms of the GNU General Public License as published by the Free
@@ -25,10 +25,11 @@
 #include "gclue-service-client.h"
 #include "gclue-service-location.h"
 #include "gclue-locator.h"
-#include "public-api/gclue-enum-types.h"
+#include "gclue-enum-types.h"
 #include "gclue-config.h"
 
 #define DEFAULT_ACCURACY_LEVEL GCLUE_ACCURACY_LEVEL_CITY
+#define DEFAULT_AGENT_STARTUP_WAIT_SECS 5
 
 static void
 gclue_service_client_client_iface_init (GClueDBusClientIface *iface);
@@ -43,16 +44,21 @@ G_DEFINE_TYPE_WITH_CODE (GClueServiceClient,
                          G_IMPLEMENT_INTERFACE (G_TYPE_INITABLE,
                                                 gclue_service_client_initable_iface_init));
 
+typedef struct _StartData StartData;
+
 struct _GClueServiceClientPrivate
 {
         GClueClientInfo *client_info;
         const char *path;
         GDBusConnection *connection;
         GClueAgent *agent_proxy;
+        StartData *pending_auth_start_data;
+        guint pending_auth_timeout_id;
 
         GClueServiceLocation *location;
         GClueServiceLocation *prev_location;
-        guint threshold;
+        guint distance_threshold;
+        guint time_threshold;
 
         GClueLocator *locator;
 
@@ -113,15 +119,15 @@ emit_location_updated (GClueServiceClient *client,
 }
 
 static gboolean
-below_threshold (GClueServiceClient *client,
-                 GClueLocation      *location)
+distance_below_threshold (GClueServiceClient *client,
+                          GClueLocation      *location)
 {
         GClueServiceClientPrivate *priv = client->priv;
         GClueLocation *cur_location;
         gdouble distance;
         gdouble threshold_km;
 
-        if (priv->threshold == 0)
+        if (priv->distance_threshold == 0)
                 return FALSE;
 
         g_object_get (priv->location,
@@ -131,13 +137,59 @@ below_threshold (GClueServiceClient *client,
                 (GEOCODE_LOCATION (cur_location), GEOCODE_LOCATION (location));
         g_object_unref (cur_location);
 
-        threshold_km = priv->threshold / 1000.0;
-        if (distance < threshold_km)
+        threshold_km = priv->distance_threshold / 1000.0;
+        if (distance < threshold_km) {
+                g_debug ("Distance from previous location is %f km and "
+                         "below threshold of %f km.",
+                         distance, threshold_km); 
                 return TRUE;
+        }
 
         return FALSE;
 }
 
+static gboolean
+time_below_threshold (GClueServiceClient *client,
+                      GClueLocation      *location)
+{
+        GClueServiceClientPrivate *priv = client->priv;
+        GClueLocation *cur_location;
+        gint64 cur_ts, ts;
+        guint64 diff_ts;
+
+        if (priv->time_threshold == 0)
+                return FALSE;
+
+        g_object_get (priv->location,
+                      "location", &cur_location,
+                      NULL);
+
+        cur_ts = geocode_location_get_timestamp
+                                (GEOCODE_LOCATION (cur_location));
+        ts = geocode_location_get_timestamp (GEOCODE_LOCATION (location));
+        diff_ts = ABS (ts - cur_ts);
+
+        g_object_unref (cur_location);
+
+        if (diff_ts < priv->time_threshold) {
+                g_debug ("Time difference between previous and new location"
+                         " is %" G_GUINT64_FORMAT " seconds and"
+                         " below threshold of %" G_GUINT32_FORMAT " seconds.",
+                         diff_ts, priv->time_threshold);
+                return TRUE;
+        }
+
+        return FALSE;
+}
+
+static gboolean
+below_threshold (GClueServiceClient *client,
+                 GClueLocation      *location)
+{
+        return (distance_below_threshold (client, location) ||
+                time_below_threshold (client, location));
+}
+
 static gboolean
 on_prev_location_timeout (gpointer user_data)
 {
@@ -211,6 +263,7 @@ start_client (GClueServiceClient *client, GClueAccuracyLevel accuracy_level)
 
         gclue_dbus_client_set_active (GCLUE_DBUS_CLIENT (client), TRUE);
         priv->locator = gclue_locator_new (accuracy_level);
+        gclue_locator_set_time_threshold (priv->locator, 0);
         g_signal_connect (priv->locator,
                           "notify::location",
                           G_CALLBACK (on_locator_location_changed),
@@ -238,7 +291,7 @@ on_agent_props_changed (GDBusProxy *agent_proxy,
         GVariant *value;
         gchar *key;
         
-        if (g_variant_n_children (changed_properties) < 0)
+        if (g_variant_n_children (changed_properties) <= 0)
                 return;
 
         g_variant_get (changed_properties, "a{sv}", &iter);
@@ -280,25 +333,28 @@ on_agent_props_changed (GDBusProxy *agent_proxy,
         g_variant_iter_free (iter);
 }
 
-typedef struct
+struct _StartData
 {
         GClueServiceClient *client;
         GDBusMethodInvocation *invocation;
-} StartData;
+        char *desktop_id;
+        GClueAccuracyLevel accuracy_level;
+};
 
 static void
 start_data_free (StartData *data)
 {
         g_object_unref (data->client);
         g_object_unref (data->invocation);
+        g_free(data->desktop_id);
         g_slice_free (StartData, data);
 }
 
 static void
-complete_start (StartData *data, GClueAccuracyLevel accuracy_level)
+complete_start (StartData *data)
 {
         GClueDBusClient *gdbus_client = GCLUE_DBUS_CLIENT (data->client);
-        start_client (data->client, accuracy_level);
+        start_client (data->client, data->accuracy_level);
 
         gclue_dbus_client_complete_start (gdbus_client,
                                           data->invocation);
@@ -313,28 +369,35 @@ on_authorize_app_ready (GObject      *source_object,
                         gpointer      user_data)
 {
         StartData *data = (StartData *) user_data;
+        GClueServiceClientPrivate *priv = data->client->priv;
         GError *error = NULL;
         gboolean authorized = FALSE;
-        GClueAccuracyLevel accuracy_level;
 
-        accuracy_level = gclue_dbus_client_get_requested_accuracy_level
-                                (GCLUE_DBUS_CLIENT (data->client));
         if (!gclue_agent_call_authorize_app_finish (GCLUE_AGENT (source_object),
                                                     &authorized,
-                                                    &accuracy_level,
+                                                    &data->accuracy_level,
                                                     res,
                                                     &error))
                 goto error_out;
 
         if (!authorized) {
-                g_set_error_literal (&error,
-                                     G_DBUS_ERROR,
-                                     G_DBUS_ERROR_ACCESS_DENIED,
-                                     "Access denied");
+                guint32 uid;
+
+                uid = gclue_client_info_get_user_id (priv->client_info);
+
+                g_set_error (&error,
+                             G_DBUS_ERROR,
+                             G_DBUS_ERROR_ACCESS_DENIED,
+                             "Agent rejected '%s' for user '%u'. Please ensure "
+                             "that '%s' has installed a valid %s.desktop file.",
+                             data->desktop_id,
+                             uid,
+                             data->desktop_id,
+                             data->desktop_id);
                 goto error_out;
         }
 
-        complete_start (data, accuracy_level);
+        complete_start (data);
 
         return;
 
@@ -343,6 +406,105 @@ error_out:
         start_data_free (data);
 }
 
+static void
+handle_post_agent_check_auth (StartData *data)
+{
+        GClueServiceClientPrivate *priv = GCLUE_SERVICE_CLIENT (data->client)->priv;
+        GClueAccuracyLevel max_accuracy;
+        GClueConfig *config;
+        GClueAppPerm app_perm;
+        guint32 uid;
+
+        uid = gclue_client_info_get_user_id (priv->client_info);
+        max_accuracy = gclue_agent_get_max_accuracy_level (priv->agent_proxy);
+
+        if (max_accuracy == 0) {
+                // Agent disabled geolocation for the user
+                g_dbus_method_invocation_return_error (data->invocation,
+                                                       G_DBUS_ERROR,
+                                                       G_DBUS_ERROR_ACCESS_DENIED,
+                                                       "Geolocation disabled for"
+                                                       " UID %u",
+                                                       uid);
+                start_data_free (data);
+                return;
+        }
+        g_debug ("requested accuracy level: %u. "
+                 "Max accuracy level allowed by agent: %u",
+                 data->accuracy_level, max_accuracy);
+        data->accuracy_level = CLAMP (data->accuracy_level, 0, max_accuracy);
+
+        config = gclue_config_get_singleton ();
+        app_perm = gclue_config_get_app_perm (config,
+                                              data->desktop_id,
+                                              priv->client_info);
+
+        if (gclue_config_is_system_component (config, data->desktop_id) ||
+            app_perm == GCLUE_APP_PERM_ALLOWED) {
+                complete_start (data);
+                return;
+        }
+
+        gclue_agent_call_authorize_app (priv->agent_proxy,
+                                        data->desktop_id,
+                                        data->accuracy_level,
+                                        NULL,
+                                        on_authorize_app_ready,
+                                        data);
+}
+
+static gboolean
+handle_pending_auth (gpointer user_data)
+{
+        GClueServiceClientPrivate *priv = GCLUE_SERVICE_CLIENT (user_data)->priv;
+        StartData *data = priv->pending_auth_start_data;
+        guint32 uid;
+
+        g_return_val_if_fail (data != NULL, G_SOURCE_REMOVE);
+
+        uid = gclue_client_info_get_user_id (priv->client_info);
+        if (priv->agent_proxy == NULL) {
+                g_dbus_method_invocation_return_error (data->invocation,
+                                                       G_DBUS_ERROR,
+                                                       G_DBUS_ERROR_ACCESS_DENIED,
+                                                       "'%s' disallowed, no agent "
+                                                       "for UID %u",
+                                                       data->desktop_id,
+                                                       uid);
+                start_data_free (data);
+        } else {
+                handle_post_agent_check_auth (data);
+        }
+
+        priv->pending_auth_timeout_id = 0;
+        priv->pending_auth_start_data = NULL;
+
+        return G_SOURCE_REMOVE;
+}
+
+static void
+set_pending_auth_timeout_enable (GClueDBusClient *client)
+{
+        GClueServiceClientPrivate *priv = GCLUE_SERVICE_CLIENT (client)->priv;
+
+        if (priv->pending_auth_timeout_id > 0)
+                return;
+        priv->pending_auth_timeout_id = g_timeout_add_seconds
+                (DEFAULT_AGENT_STARTUP_WAIT_SECS, handle_pending_auth, client);
+}
+
+static void
+set_pending_auth_timeout_disable (GClueDBusClient *client)
+{
+        GClueServiceClientPrivate *priv = GCLUE_SERVICE_CLIENT (client)->priv;
+
+        if (priv->pending_auth_timeout_id == 0)
+                return;
+
+        g_source_remove (priv->pending_auth_timeout_id);
+        priv->pending_auth_timeout_id = 0;
+}
+
 static gboolean
 gclue_service_client_handle_start (GClueDBusClient       *client,
                                    GDBusMethodInvocation *invocation)
@@ -351,20 +513,26 @@ gclue_service_client_handle_start (GClueDBusClient       *client,
         GClueConfig *config;
         StartData *data;
         const char *desktop_id;
-        GClueAccuracyLevel accuracy_level, max_accuracy;
         GClueAppPerm app_perm;
         guint32 uid;
 
-        if (priv->locator != NULL)
+        if (priv->locator != NULL) {
                 /* Already started */
+                gclue_dbus_client_complete_start (client, invocation);
+
                 return TRUE;
+        }
+
+        desktop_id = gclue_client_info_get_xdg_id (priv->client_info);
+        if (desktop_id == NULL)
+                /* Non-xdg app */
+                desktop_id = gclue_dbus_client_get_desktop_id (client);
 
-        desktop_id = gclue_dbus_client_get_desktop_id (client);
         if (desktop_id == NULL) {
-                g_dbus_method_invocation_return_error (invocation,
-                                                       G_DBUS_ERROR,
-                                                       G_DBUS_ERROR_ACCESS_DENIED,
-                                                       "'DesktopId' property must be set");
+                g_dbus_method_invocation_return_error_literal (invocation,
+                                                               G_DBUS_ERROR,
+                                                               G_DBUS_ERROR_ACCESS_DENIED,
+                                                               "'DesktopId' property must be set");
                 return TRUE;
         }
 
@@ -385,49 +553,32 @@ gclue_service_client_handle_start (GClueDBusClient       *client,
         }
 
         data = g_slice_new (StartData);
-        data->client = g_object_ref (client);
+        data->client = g_object_ref (GCLUE_SERVICE_CLIENT (client));
         data->invocation =  g_object_ref (invocation);
+        data->desktop_id =  g_strdup (desktop_id);
 
-        accuracy_level = gclue_dbus_client_get_requested_accuracy_level (client);
-        accuracy_level = CLAMP (accuracy_level,
-                                GCLUE_ACCURACY_LEVEL_COUNTRY,
-                                GCLUE_ACCURACY_LEVEL_EXACT);
-
-        /* No agent == No authorization needed */
-        if (priv->agent_proxy == NULL ||
-            gclue_config_is_system_component (config, desktop_id) ||
-            app_perm == GCLUE_APP_PERM_ALLOWED) {
-                complete_start (data, accuracy_level);
+        data->accuracy_level = gclue_dbus_client_get_requested_accuracy_level (client);
+        data->accuracy_level = CLAMP (data->accuracy_level,
+                                      GCLUE_ACCURACY_LEVEL_COUNTRY,
+                                      GCLUE_ACCURACY_LEVEL_EXACT);
 
+        /* No agent == No authorization */
+        if (priv->agent_proxy == NULL) {
+                /* Already a pending Start()? Denied! */
+                if (priv->pending_auth_start_data) {
+                        g_dbus_method_invocation_return_error_literal
+                                (invocation,
+                                 G_DBUS_ERROR,
+                                 G_DBUS_ERROR_ACCESS_DENIED,
+                                 "An authorization request is already pending");
+                } else {
+                        priv->pending_auth_start_data = data;
+                        set_pending_auth_timeout_enable (client);
+                }
                 return TRUE;
         }
 
-        if (priv->agent_proxy != NULL)
-                max_accuracy = gclue_agent_get_max_accuracy_level (priv->agent_proxy);
-        else
-                max_accuracy = GCLUE_ACCURACY_LEVEL_EXACT;
-
-        if (max_accuracy == 0) {
-                g_dbus_method_invocation_return_error (invocation,
-                                                       G_DBUS_ERROR,
-                                                       G_DBUS_ERROR_ACCESS_DENIED,
-                                                       "Geolocation disabled for"
-                                                       " UID %u",
-                                                       uid);
-                start_data_free (data);
-                return TRUE;
-        }
-        g_debug ("requested accuracy level: %u. "
-                 "Accuracy level allowed by agent: %u",
-                 accuracy_level, max_accuracy);
-        accuracy_level = CLAMP (accuracy_level, 0, max_accuracy);
-
-        gclue_agent_call_authorize_app (priv->agent_proxy,
-                                        desktop_id,
-                                        accuracy_level,
-                                        NULL,
-                                        on_authorize_app_ready,
-                                        data);
+        handle_post_agent_check_auth (data);
 
         return TRUE;
 }
@@ -450,6 +601,8 @@ gclue_service_client_finalize (GObject *object)
 
         g_clear_pointer (&priv->path, g_free);
         g_clear_object (&priv->connection);
+        set_pending_auth_timeout_disable (GCLUE_DBUS_CLIENT (object));
+        g_clear_pointer (&priv->pending_auth_start_data, start_data_free);
         if (priv->agent_proxy != NULL)
                 g_signal_handlers_disconnect_by_func
                                 (priv->agent_proxy,
@@ -523,6 +676,8 @@ gclue_service_client_set_property (GObject      *object,
                                           "g-properties-changed",
                                           G_CALLBACK (on_agent_props_changed),
                                           object);
+                if (client->priv->pending_auth_start_data != NULL)
+                        handle_pending_auth (client);
                 break;
 
         default:
@@ -545,10 +700,10 @@ gclue_service_client_handle_method_call (GDBusConnection       *connection,
         GDBusInterfaceVTable *skeleton_vtable;
 
         if (!gclue_client_info_check_bus_name (priv->client_info, sender)) {
-                g_dbus_method_invocation_return_error (invocation,
-                                                       G_DBUS_ERROR,
-                                                       G_DBUS_ERROR_ACCESS_DENIED,
-                                                       "Access denied");
+                g_dbus_method_invocation_return_error_literal (invocation,
+                                                               G_DBUS_ERROR,
+                                                               G_DBUS_ERROR_ACCESS_DENIED,
+                                                               "Access denied");
                 return;
         }
 
@@ -631,9 +786,17 @@ gclue_service_client_handle_set_property (GDBusConnection *connection,
                                              error,
                                              user_data);
         if (ret && strcmp (property_name, "DistanceThreshold") == 0) {
-                priv->threshold = gclue_dbus_client_get_distance_threshold
+                priv->distance_threshold = gclue_dbus_client_get_distance_threshold
                         (client);
-                g_debug ("New distance threshold: %u", priv->threshold);
+                g_debug ("New distance threshold: %u", priv->distance_threshold);
+        } else if (ret && strcmp (property_name, "TimeThreshold") == 0) {
+                priv->time_threshold = gclue_dbus_client_get_time_threshold
+                        (client);
+                gclue_locator_set_time_threshold (priv->locator,
+                                                  priv->time_threshold);
+                g_debug ("%s: New time-threshold:  %u",
+                         G_OBJECT_TYPE_NAME (client),
+                         priv->time_threshold);
         }
 
         return ret;
