@@ -1,7 +1,7 @@
 /* vim: set et ts=8 sw=8: */
 /* gclue-service-manager.c
  *
- * Copyright (C) 2013 Red Hat, Inc.
+ * Copyright 2013 Red Hat, Inc.
  *
  * Geoclue is free software; you can redistribute it and/or modify it under
  * the terms of the GNU General Public License as published by the Free
@@ -173,10 +173,10 @@ complete_get_client (OnClientInfoNewReadyData *data)
         goto out;
 
 error_out:
-        g_dbus_method_invocation_return_error (data->invocation,
-                                               G_DBUS_ERROR,
-                                               G_DBUS_ERROR_FAILED,
-                                               "%s", error->message);
+        g_dbus_method_invocation_return_error_literal (data->invocation,
+                                                       G_DBUS_ERROR,
+                                                       G_DBUS_ERROR_FAILED,
+                                                       error->message);
 out:
         g_clear_error (&error);
         g_clear_object (&info);
@@ -201,10 +201,10 @@ on_client_info_new_ready (GObject      *source_object,
 
         info = gclue_client_info_new_finish (res, &error);
         if (info == NULL) {
-                g_dbus_method_invocation_return_error (data->invocation,
-                                                       G_DBUS_ERROR,
-                                                       G_DBUS_ERROR_FAILED,
-                                                       "%s", error->message);
+                g_dbus_method_invocation_return_error_literal (data->invocation,
+                                                               G_DBUS_ERROR,
+                                                               G_DBUS_ERROR_FAILED,
+                                                               error->message);
                 g_error_free (error);
                 g_slice_free (OnClientInfoNewReadyData, data);
 
@@ -322,10 +322,10 @@ on_agent_proxy_ready (GObject      *source_object,
         goto out;
 
 error_out:
-        g_dbus_method_invocation_return_error (data->invocation,
-                                               G_DBUS_ERROR,
-                                               G_DBUS_ERROR_FAILED,
-                                               "%s", error->message);
+        g_dbus_method_invocation_return_error_literal (data->invocation,
+                                                       G_DBUS_ERROR,
+                                                       G_DBUS_ERROR_FAILED,
+                                                       error->message);
 out:
         g_clear_error (&error);
         add_agent_data_free (data);
@@ -341,27 +341,31 @@ on_agent_info_new_ready (GObject      *source_object,
         AddAgentData *data = (AddAgentData *) user_data;
         GError *error = NULL;
         GClueConfig *config;
+        const char *xdg_id;
 
         data->info = gclue_client_info_new_finish (res, &error);
         if (data->info == NULL) {
-                g_dbus_method_invocation_return_error (data->invocation,
-                                                       G_DBUS_ERROR,
-                                                       G_DBUS_ERROR_FAILED,
-                                                       "%s", error->message);
+                g_dbus_method_invocation_return_error_literal (data->invocation,
+                                                               G_DBUS_ERROR,
+                                                               G_DBUS_ERROR_FAILED,
+                                                               error->message);
                 g_error_free (error);
                 add_agent_data_free (data);
 
                 return;
         }
 
+        xdg_id = gclue_client_info_get_xdg_id (data->info);
         config = gclue_config_get_singleton ();
-        if (!gclue_config_is_agent_allowed (config,
+        if (xdg_id != NULL ||
+            !gclue_config_is_agent_allowed (config,
                                             data->desktop_id,
                                             data->info)) {
                 g_dbus_method_invocation_return_error (data->invocation,
                                                        G_DBUS_ERROR,
                                                        G_DBUS_ERROR_ACCESS_DENIED,
-                                                       "Not whitelisted");
+                                                       "%s not allowed to act as agent",
+                                                       data->desktop_id);
                 add_agent_data_free (data);
 
                 return;
